'''Game main module.

Contains the entry point used by the run_game.py script.

Feel free to put all your game code here, or in other modules in this "lib"
directory.
'''
import data
import pygame
from pygame.locals import *

pygame.mixer.init()
import math, time, random, os

# globals
_screen = None
_scr_width = 800
_scr_height = 600

_world = None
_ship1 = None
_ship2 = None

# options
SHIP_COLLISION = False

KILLS_TO_WIN = 5
_min_kills = 5
_max_kills = 50

SOUND = True

## world coordinates are specified with regularly named variables x,y, length etc
## Screen coords use s_x,s_y s_length, etc

def load_image(name, transparency = False):
    im = pygame.image.load(data.filepath(name))
    if transparency:
        corner = im.get_at((0, 0))
        im.set_colorkey(corner)
    return im.convert()

def cls():
    _screen.fill(data.black)

def putText(text, size, color, x, y):
    global _screen
    font = pygame.font.Font(None, size)
    font_im = font.render(text, True, color)
    _screen.blit(font_im, (x, y))
    
class Actor(object):
    LIVE = 1
    SPAWN = 2
    TWIST = 3

    def __init__(self):
        self.state = Actor.LIVE

    def get_image(self):
        return self.image
    
    def get_rect(self):
        width, height = self.image.get_size()
        r = pygame.Rect((self.x - (width / 2),
                         self.y - (height / 2),
                         width, height))
        r.inflate(-2, -2)
        return r

    def handle_collision_with(self, other_actor):
        pass

    def tick(self):
        self.x += self.vx
        self.y += self.vy
            

class Animation(object):

    def __init__(self, image_base, num_images, frame_delay = 1):
        self.images = []
        self.frame_delay = frame_delay
        self.frame_countdown = self.frame_delay
        for i in range(num_images):
            im = load_image(image_base + str(i + 1) + ".bmp", True)
            self.images.append(im)

        self.current_frame = 0

    def get_next_frame(self):
        if self.frame_countdown > 0:
            self.frame_countdown -= 1

        if self.frame_countdown == 0:
            self.current_frame += 1
            if self.current_frame > (len(self.images) - 1):
                self.current_frame = 0
            self.frame_countdown = self.frame_delay

        return self.images[self.current_frame]


class World(object):
    def __init__(self):
        self.actors = []
        self.props = []

    def add(self, actor):
        self.actors.append(actor)
        
    def remove(self, actor):
        self.actors.remove(actor)

    def add_prop(self, prop):
        self.props.append(prop)


    def tick(self):
        for actor in self.actors:
            actor.tick()

        for prop in self.props:
            prop.tick()
            
        for actor1 in self.actors:
            for actor2 in self.actors:
                if not (actor1 is actor2) and actor1.state == Actor.LIVE and actor2.state == Actor.LIVE:
                    if actor1.get_rect().colliderect(actor2.get_rect()):
                        actor1.handle_collision_with(actor2)
    
            

    def render(self):
        for actor in self.actors:
            if actor.state == Actor.LIVE or actor.state == Actor.TWIST:
                _screen.blit(actor.get_image(),
                             (actor.x - actor.get_image().get_width() / 2,
                              actor.y - actor.get_image().get_height() / 2))
            
class BlackHole(Actor):

    def __init__(self, x, y, partner, vx = 0, vy = 0):
        super(BlackHole, self).__init__()
        self.x = x
        self.y = y
        self.vx = vx
        self.vy = vy
        self.animation = Animation("black_hole_", 6, frame_delay = 10)
        self.image = self.animation.images[0]
        self.partner = partner

    def get_image(self):
        return self.animation.get_next_frame()
    
    def tick(self):
        super(BlackHole, self).tick()
        # Wrapping
        if self.x < -self.image.get_width():
            self.x = _scr_width
        if self.x > _scr_width:
            self.x = -self.image.get_width()
        if self.y < -self.image.get_height():
            self.y = _scr_height
        if self.y > _scr_height:
            self.y = -self.image.get_height()

class WhiteHole(Actor):

    def __init__(self, x, y, vx = 0, vy = 0):
        super(WhiteHole, self).__init__()
        self.x = x
        self.y = y
        self.vx = vx
        self.vy = vy
        self.animation = Animation("white_hole_", 6, frame_delay = 10)
        self.image = self.animation.images[0]
        
    def get_image(self):
        return self.animation.get_next_frame()
    
    def tick(self):
        super(WhiteHole, self).tick()
        # Wrapping
        if self.x < -self.image.get_width():
            self.x = _scr_width
        if self.x > _scr_width:
            self.x = -self.image.get_width()
        if self.y < -self.image.get_height():
            self.y = _scr_height
        if self.y > _scr_height:
            self.y = -self.image.get_height()

class Twister(object):

    def __init__(self, x, y, black_hole, white_hole):
        self.x = x
        self.y = y
        self.arm_length = 600
        self.angle = random.randrange(360)
        self.angular_speed = 0.1
        velocity_components = [-2, -1, 1, 2]
        self.vx = random.choice(velocity_components)
        self.vy = random.choice(velocity_components)
        self.black_hole = black_hole
        self.white_hole = white_hole
        
    def tick(self):
        self.x += self.vx
        self.y += self.vy
        self.angle = (self.angle + self.angular_speed) % 360.0

        self.black_hole.x = self.x + self.arm_length / 2 * math.cos(math.radians(self.angle))
        self.black_hole.y = self.y - self.arm_length / 2 * math.sin(math.radians(self.angle))

        self.white_hole.x = self.x + self.arm_length / 2 * math.cos(math.radians(self.angle + 180))
        self.white_hole.y = self.y - self.arm_length / 2 * math.sin(math.radians(self.angle + 180))

        self.black_hole.x = self.black_hole.x % _scr_width
        self.black_hole.y = self.black_hole.y % _scr_height
        self.white_hole.x = self.white_hole.x % _scr_width
        self.white_hole.y = self.white_hole.y % _scr_height

        
class Ship(Actor):
    """A spaceship."""
    ROTATION_SPEED = 3
    THRUST = 1
    BOOST_MULTIPLIER = 3
    BOOST_PENALTY = 1
    MAX_X_VELOCITY = 1.5
    MAX_Y_VELOCITY = 1.5
    SPAWN_TIME = 30
    SPAWN_BUFFER = 200
    
    def __init__(self, x, y, player = 1):
        super(Ship, self).__init__()
        self.x = x
        self.y = y
        self.player = player
        self.animations = [Animation("ship" + str(player) + "_", 1),
                           Animation("ship" + str(player) + "_thrust_", 2,
                                     frame_delay = 5),
                           Animation("ship" + str(player) + "_boost_", 2,
                                     frame_delay = 5)]
        self.current_animation = self.animations[0]
        self.width, self.height = self.animations[0].images[0].get_size()
        self.angle = 0
        self.vx = 0
        self.vy = 0
        self.kills = 0
        self.shields = 100
        
        self.twist_countdown = 30
        self.zoom_factor = 1.0
        
        self.thrusting = False
        self.boosting = False

    def get_rect(self):
        r = pygame.Rect((self.x - (self.width / 2),
                         self.y - (self.height / 2),
                         self.width, self.height))
        r.inflate(-2, -2)
        return r

    def thrust(self):
        ax = Ship.THRUST * math.cos(math.radians(self.angle + 90))
        ay = Ship.THRUST * math.sin(math.radians(self.angle - 90))
        if abs(self.vx + ax) <= Ship.MAX_X_VELOCITY:
            self.vx += ax
        if abs(self.vy + ay) <= Ship.MAX_Y_VELOCITY:
            self.vy += ay
        self.thrusting = True

    def boost(self):
        if self.shields > Ship.BOOST_PENALTY * 2:
            ax = Ship.THRUST * math.cos(math.radians(self.angle + 90))
            ay = Ship.THRUST * math.sin(math.radians(self.angle - 90))
            ax *= Ship.BOOST_MULTIPLIER
            ay *= Ship.BOOST_MULTIPLIER
            if abs(self.vx + ax) <= Ship.MAX_X_VELOCITY * Ship.BOOST_MULTIPLIER:
                self.vx += ax
            if abs(self.vy + ay) <= Ship.MAX_Y_VELOCITY * Ship.BOOST_MULTIPLIER:
                self.vy += ay
            self.boosting = True
            self.shields -= Ship.BOOST_PENALTY
        
    def back_thrust(self):
        ax = Ship.THRUST * math.cos(math.radians(self.angle + 90))
        ay = Ship.THRUST * math.sin(math.radians(self.angle - 90))
        if abs(self.vx - ax) <= Ship.MAX_X_VELOCITY:
            self.vx -= ax
        if abs(self.vy - ay) <= Ship.MAX_Y_VELOCITY:
            self.vy -= ay
        
        
    def get_image(self):
        if self.state == Actor.LIVE:
            im = pygame.transform.rotate(self.current_animation.get_next_frame(),
                                         self.angle)
        elif self.state == Actor.TWIST:
            im = self.current_animation.get_next_frame()
            colorkey = im.get_colorkey()
            im = pygame.transform.rotozoom(im, self.angle, self.zoom_factor)
            im.set_colorkey(colorkey)
        return im
    
    def turn_left(self):
        self.angle = (self.angle + Ship.ROTATION_SPEED) % 360

    def turn_right(self):
        self.angle = (self.angle - Ship.ROTATION_SPEED) % 360

    def fire(self):
        global _world
        missile = Missile(self.x, self.y, self.angle + 90)
        _world.add(missile)

    def tick(self):
        if self.state == Actor.SPAWN:
            self.spawn_delay -= 1
            if self.spawn_delay <= 0:
                self.state = Actor.LIVE
        elif self.state == Actor.LIVE:
            super(Ship, self).tick()
            # Wrapping
            if self.x < -self.width:
                self.x = _scr_width
            if self.x > _scr_width:
                self.x = -self.width
            if self.y < -self.height:
                self.y = _scr_height
            if self.y > _scr_height:
                self.y = -self.height

            self.shields += 0.05
            if self.shields >= 100:
                self.shields = 100
            self.vx *= 0.999
            self.vy *= 0.999

            if self.thrusting:
                self.current_animation = self.animations[1]
            if self.boosting:
                self.current_animation = self.animations[2]

            if not self.thrusting and not self.boosting:
                self.current_animation = self.animations[0]

            self.thrusting = False
            self.boosting = False
        elif self.state == Actor.TWIST:
            self.zoom_factor -= 0.03
            self.twist_countdown -= 1
            if self.twist_countdown > 0:
                self.angle = (self.angle - 40) % 360
            else:
                self.twist_countdown = 30
                self.zoom_factor = 1.0
                self.state = Actor.LIVE
                self.x = self.black_hole.partner.x
                self.y = self.black_hole.partner.y
            
        if self.shields < 0:
            self.destroy()

    def handle_collision_with(self, other_actor):
        if isinstance(other_actor, Missile):
            self.shields -= 10
            if self.shields < 0:
                self.destroy()
            other_actor.destroy()
        elif isinstance(other_actor, Ship):
            if SHIP_COLLISION:
                self.shields = 0
                self.destroy()
                other_actor.destroy()
        elif isinstance(other_actor, BlackHole):
            self.state = Actor.TWIST
            self.black_hole = other_actor
    
    def destroy(self):
        global _world, _ship1, _ship2
        exp = Explosion(self.x, self.y)
        _world.add(exp)
        self.state = Actor.SPAWN
        self.spawn_delay = Ship.SPAWN_TIME

        while True:
            self.x = random.randint(100, _scr_width - 100)
            self.y = random.randint(100, _scr_height - 100)
            if self.player == 1:
                dist_x = abs(_ship2.x - self.x)
                dist_y = abs(_ship2.y - self.y)
            elif self.player == 2:
                dist_x = abs(_ship1.x - self.x)
                dist_y = abs(_ship1.y - self.y)
            if dist_x > Ship.SPAWN_BUFFER and dist_y > Ship.SPAWN_BUFFER:
                break
                
        self.angle = random.randint(0, 359)
        self.shields = 100
        
        if self.player == 1:
            _ship2.kills += 1
        elif self.player == 2:
            _ship1.kills += 1
    
class Missile(Actor):
    VELOCITY = 5

    def __init__(self, x, y, angle):
        super(Missile, self).__init__()
        self.x = x
        self.y = y
        self.vx = math.cos(math.radians(angle)) * Missile.VELOCITY
        self.vy = -math.sin(math.radians(angle)) * Missile.VELOCITY
        self.angle = angle
        self.x += self.vx * 10
        self.y += self.vy * 10
        self.lifetime = 60
        self.image = load_image("missile.bmp", True)

    def get_image(self):
        im = pygame.transform.rotate(self.image, self.angle - 90)
        return im
    
    def tick(self):
        self.lifetime -= 1
        if self.lifetime == 0:
            self.destroy()

        super(Missile, self).tick()

    def handle_collision_with(self, other_actor):
        if isinstance(other_actor, BlackHole):
            self.x = other_actor.partner.x
            self.y = other_actor.partner.y
            
    def destroy(self):
        global _world
        _world.remove(self)
    

class Explosion(Actor):
    """
    Although we have an Animation class, the Animation class is meant
    more for animated sprites, rather than pure animation.
    """
    FRAME_DELAY = 2
    sound = pygame.mixer.Sound(data.filepath("explosion.wav"))
    images = None
        
    def __init__(self, x, y):
        if not Explosion.images:
            Explosion.images = []
            for i in range(1, 15):
                im = load_image("ex" + os.sep + "ex" + str(i) + ".bmp", True)
                Explosion.images.append(im)
            
        super(Explosion, self).__init__()
        self.x = x
        self.y = y
        self.images = Explosion.images
        self.frame = 0
        self.frame_delay = Explosion.FRAME_DELAY
        self.image = self.images[self.frame]
        if SOUND:
            Explosion.sound.play()

    def get_next_frame(self):
        if self.frame >= 13:
            self.destroy()
        else:
            self.frame += 1

    def tick(self):
        self.image = self.images[self.frame]
        self.frame_delay -= 1
        if self.frame_delay == 0:
            self.get_next_frame()
            self.frame_delay = Explosion.FRAME_DELAY

    def get_image(self):
        return self.image
    
    def destroy(self):
        global _world
        _world.remove(self)
        

def init_screen():
    global _screen
    pygame.init()
    _world = World()
    _screen = pygame.display.set_mode((_scr_width, _scr_height))
    pygame.display.set_caption("Nemesis: A Game Of Twisted Space/Time")


def menu():
    global _screen
    cls()
    background = load_image("twisted_space.jpg")
    background_x = 0
    background_y = 0
    background_w, background_h = background.get_size()

    menu_choice = 0
    choices = ("Deathmatch",
               "Instructions",
               "Options",
               "Quit")
    
    while True:
        for event in pygame.event.get():

            if event.type == QUIT:
                return len(choices) - 1

            elif event.type == KEYDOWN:
                if event.key == K_RETURN:
                    return menu_choice
                elif event.key == K_UP:
                    menu_choice -= 1
                    if menu_choice < 0:
                        menu_choice = len(choices) - 1
                elif event.key == K_DOWN:
                    menu_choice += 1
                    if menu_choice > (len(choices) - 1):
                        menu_choice = 0

        cls()
        _screen.blit(background, (background_x, background_y))
        _screen.blit(background, (background_x + background_w, background_y))
        _screen.blit(background, (background_x, background_y + background_h))
        _screen.blit(background, (background_x + background_w,
                                  background_y + background_h))
        background_x -= 1
        background_y -= 1
        if background_x < -background_w:
            background_x = 0
        if background_y < -background_h:
            background_y = 0
        
        putText("Nemesis", 60, data.red, 10, 30)
        y = 150
        for choice in choices:
            if choices[menu_choice] == choice:
                putText("[>]", 40, data.yellow, 5, y)
            putText(choice, 40, data.white, 50, y)
            y += 60

        pygame.display.flip() 

def display_instructions():
    cls()

    introduction = ("NEMESIS INSTRUCTIONS",
                    "",
                    "Welcome to Nemesis, a space shooter with a twist!",
                    "",
                    "In this game, you must kill off your opponent. The",
                    "catch is that the black holes orbiting the arena will",
                    "teleport anything, including you, to a corresponding",
                    "white hole, should anything come into contact with one!",
                    "")

    controls_1 = ("PLAYER 1 CONTROLS",
                  "      A - Rotate left",
                  "      D - Rotate right",
                  "      S - Back thrust",
                  "      W - Forward thrust",
                  "      W + Left Ctrl - Boost",
                  "      Space - Fire")

    controls_2 = ("PLAYER 2 CONTROLS",
                  "      Left arrow - Rotate left",
                  "      Right arrow - Rotate right",
                  "      Down arrow - Back thrust",
                  "      Up arrow - Forward thrust",
                  "      Up arrow + Num 0 - Boost",
                  "      Num Enter - Fire")

    notes = ("When hit, your shielding will drop. Shielding regenerates",
             "over time, though not nearly as fast as hits will drain it,",
             "so watch your shielding carefully!",
             "",
             "If you run into your opponent while ship collision is on",
             "(see options), one point will be awarded to both players.",
             "",
             "Boosting diverts shielding power, so use sparingly!",
             "Even though the ship will stop diverting power when low,",
             "your enemy may finish you off before your power regenerates!")
    
    while True:
        for event in pygame.event.get():

            if event.type == QUIT:
                shutdown()

            if event.type == KEYDOWN:
                return

        cls()
        y = 20
        for text in introduction:
            putText(text, 25, data.white, 20, y)
            y += 20

        y = 200
        for text in controls_1:
            putText(text, 25, data.yellow, 20, y)
            y += 20

        y = 200
        for text in controls_2:
            putText(text, 25, data.red, 400, y)
            y += 20

        y = 360
        for text in notes:
            putText(text, 25, data.white, 20, y)
            y += 20
            
        putText("Press any key to return to the main menu", 30,
                data.white, 20, 570)
        pygame.display.flip()

def options():
    global SHIP_COLLISION, KILLS_TO_WIN, SOUND
    cls()

    menu_choice = 0
    choices = [["Ship collision: ", SHIP_COLLISION],
               ["Sound: ", SOUND],
               ["Kills to win: ", KILLS_TO_WIN]]
    
    while True:
        for event in pygame.event.get():

            if event.type == QUIT:
                shutdown()

            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    return
                elif event.key == K_RIGHT:
                    if menu_choice == 0:
                        SHIP_COLLISION = not SHIP_COLLISION
                        choices[0][1] = SHIP_COLLISION
                    elif menu_choice == 1:
                        SOUND = not SOUND
                        choices[1][1] = SOUND
                    elif menu_choice == 2:
                        KILLS_TO_WIN += 5
                        if KILLS_TO_WIN > _max_kills:
                            KILLS_TO_WIN = _min_kills
                        choices[2][1] = KILLS_TO_WIN
                elif event.key == K_LEFT:
                    if menu_choice == 0:
                        SHIP_COLLISION = not SHIP_COLLISION
                        choices[0][1] = SHIP_COLLISION
                    elif menu_choice == 1:
                        SOUND = not SOUND
                        choices[1][1] = SOUND
                    elif menu_choice == 2:
                        KILLS_TO_WIN -= 5
                        if KILLS_TO_WIN < _min_kills:
                            KILLS_TO_WIN = _max_kills
                        choices[2][1] = KILLS_TO_WIN
                        
                elif event.key == K_UP:
                    menu_choice -= 1
                    if menu_choice < 0:
                        menu_choice = len(choices) - 1
                elif event.key == K_DOWN:
                    menu_choice += 1
                    if menu_choice > (len(choices) - 1):
                        menu_choice = 0

            cls()
            putText("OPTIONS", 50, data.red, 20, 20)
            y = 80
            for choice, value in choices:
                if choices[menu_choice][0] == choice:
                    putText("[>]", 40, data.yellow, 5, y)
                putText(choice + str(value), 40, data.white, 50, y)
                y += 40
            putText("Use left and right arrow keys to select", 30, data.white,
                    20, 560)
                
        pygame.display.flip()
        
def init_objects():
    global _world, _ship1, _ship2
    _world = World()
    white_hole1 = WhiteHole(0, 0)
    black_hole1 = BlackHole(0, 0, white_hole1)
    white_hole2 = WhiteHole(0, 0)
    black_hole2 = BlackHole(0, 0, white_hole2)

    vx = random.randint(-3, 3)
    vy = random.randint(-3, 3)
    white_hole3 = WhiteHole(100, 100, vx, vy)
    black_hole3 = BlackHole(200, 200, white_hole3, -vx, -vy)
    
    _world.add(white_hole1)
    _world.add(black_hole1)
    _world.add(white_hole2)
    _world.add(black_hole2)
    _world.add(white_hole3)
    _world.add(black_hole3)
    twister_arm1 = Twister(300, 300, white_hole1, black_hole1)
    twister_arm2 = Twister(400, 450, white_hole2, black_hole2)
    _world.add_prop(twister_arm1)
    _world.add_prop(twister_arm2)
    
    _ship1 = Ship(100, 300, 1)
    _ship2 = Ship(600, 300, 2)
    _world.add(_ship1)
    _world.add(_ship2)
    

def event_loop():
    background = load_image("twisted_space.jpg")
    background_x = 0
    background_y = 0
    background_w, background_h = background.get_size()
    
    clock = pygame.time.Clock()
    while True:
        clock.tick(60)
        
        events = pygame.event.get()
        for evt in events:
            if (evt.type == QUIT):
                shutdown()
                return
            
            if (evt.type == KEYDOWN and evt.key == K_ESCAPE):
                return

            # Firing for players
            if evt.type == KEYDOWN:
                if evt.key == K_SPACE:
                    _ship1.fire()
                elif evt.key == K_KP_ENTER:
                    _ship2.fire()

        # Movement for player 1
        if pygame.key.get_pressed()[K_a]:
            _ship1.turn_left()
        if pygame.key.get_pressed()[K_d]:
            _ship1.turn_right()
        if pygame.key.get_pressed()[K_w]:
            if pygame.key.get_pressed()[K_LCTRL]:
                _ship1.boost()
            else:
                _ship1.thrust()
        if pygame.key.get_pressed()[K_s]:
            _ship1.back_thrust()

        # Movement for player 2
        if pygame.key.get_pressed()[K_LEFT]:
            _ship2.turn_left()
        if pygame.key.get_pressed()[K_RIGHT]:
            _ship2.turn_right()
        if pygame.key.get_pressed()[K_UP]:
            if pygame.key.get_pressed()[K_KP0]:
                _ship2.boost()
            else:
                _ship2.thrust()
        if pygame.key.get_pressed()[K_DOWN]:
            _ship2.back_thrust()
            
        _world.tick()

        cls()

        # Background scrolling
        _screen.blit(background, (background_x, background_y))
        _screen.blit(background, (background_x + background_w, background_y))
        _screen.blit(background, (background_x, background_y + background_h))
        _screen.blit(background, (background_x + background_w,
                                  background_y + background_h))
        background_x -= 1
        background_y -= 1
        if background_x < -background_w:
            background_x = 0
        if background_y < -background_h:
            background_y = 0
            
        putText("Player 1: " + str(_ship1.kills), 25, data.yellow, 10, 10)
        putText("Shields: " + str(_ship1.shields) + "%", 25, data.yellow, 10, 30)
        putText("Player 2: " + str(_ship2.kills), 25, data.red, 680, 10)
        putText("Shields: " + str(_ship2.shields) + "%", 25, data.red, 680, 30)
        
        _world.render()

        if _ship1.kills >= KILLS_TO_WIN and _ship2.kills >= KILLS_TO_WIN:
            victory(0)
            return
        elif _ship1.kills >= KILLS_TO_WIN:
            victory(1)
            return
        elif _ship2.kills >= KILLS_TO_WIN:
            victory(2)
            return
  
        pygame.display.flip()

def victory(player):
    cls()

    if player >= 1:
        putText("Player " + str(player) + " wins!", 60, data.red, 20, 50)
        putText("- with a score of " + str(_ship1.kills) + " to " + str(_ship2.kills),
                60, data.yellow, 20, 100)
    else:
        putText("It's a draw!", 60, data.red, 20, 50)

    putText("Press Esc to return to the main menu", 30, data.white, 20, 560)
    pygame.display.flip()
    while True:
        for event in pygame.event.get():
            if event.type == QUIT:
                shutdown()
                return
            elif event.type == KEYDOWN and event.key == K_ESCAPE:
                return
            

def shutdown():
    pygame.quit()
    
    
def main():
    init_screen()
    while True:
        choice = menu()

        if choice == 0:
            init_objects()
            event_loop()
        elif choice == 1:
            display_instructions()
        elif choice == 2:
            options()
        else:
            break

    shutdown()

if __name__ == "__main__":
    print "Run this game using run_game.py"
    raw_input()
    
